{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Question 73 - LC 117. Populating Next Right Pointers in Each Node II\n",
    "\n",
    "Given a binary tree:\n",
    "\n",
    "```c++\n",
    "struct Node {\n",
    "  int val;\n",
    "  Node *left;\n",
    "  Node *right;\n",
    "  Node *next;\n",
    "}\n",
    "```\n",
    "\n",
    "```python\n",
    "# Definition for a Node.\n",
    "class Node:\n",
    "    def __init__(self, val: int = 0, left: 'Node' = None, right: 'Node' = None, next: 'Node' = None):\n",
    "        self.val = val\n",
    "        self.left = left\n",
    "        self.right = right\n",
    "        self.next = next\n",
    "```\n",
    "\n",
    "Populate each next pointer to point to its next right node. If there is no next right node, the next pointer should be set to `NULL`.\n",
    "\n",
    "Initially, all next pointers are set to `NULL`.\n",
    "\n",
    "**Example 1:**\n",
    "\n",
    "    Input: root = [1,2,3,4,5,null,7]\n",
    "    Output: [1,#,2,3,#,4,5,7,#]\n",
    "    Explanation: Given the above binary tree (Figure A), your function should populate each next pointer to point to its next right node, just like in Figure B. The serialized output is in level order as connected by the next pointers, with '#' signifying the end of each level.\n",
    "\n",
    "**Example 2:**\n",
    "\n",
    "    Input: root = []\n",
    "    Output: []\n",
    "\n",
    "**Constraints:**\n",
    "\n",
    "- The number of nodes in the tree is in the range [0, 6000].\n",
    "- -100 <= Node.val <= 100\n",
    "\n",
    "**Follow-up:**\n",
    "\n",
    "- You may only use constant extra space.\n",
    "- The recursive approach is fine. You may assume implicit stack space does not count as extra space for this problem.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [],
   "source": [
    "from __future__ import annotations\n",
    "from typing import Optional\n",
    "from collections import deque"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [],
   "source": [
    "# Definition for a binary tree node.\n",
    "class Node:\n",
    "    def __init__(self, val=0, left=None, right=None, next=None):\n",
    "        self.val = val\n",
    "        self.left = left\n",
    "        self.right = right\n",
    "        self.next = next\n",
    "\n",
    "    @staticmethod\n",
    "    def from_list(values: list[Optional[int]]) -> Optional[Node]:\n",
    "        \"\"\"\n",
    "        Constructs a binary tree from a list of values using level-order traversal (BFS).\n",
    "\n",
    "        Parameters:\n",
    "        - values (list): The list of values to construct the binary tree from.\n",
    "\n",
    "        Returns:\n",
    "        TreeNode: The root of the constructed binary tree.\n",
    "        \"\"\"\n",
    "        if not values:\n",
    "            return None\n",
    "\n",
    "        root = Node(val=values[0])\n",
    "        queue = deque([root])\n",
    "        i = 1\n",
    "\n",
    "        while queue and i < len(values):\n",
    "            current_node = queue.popleft()\n",
    "\n",
    "            if values[i] is not None:\n",
    "                current_node.left = Node(val=values[i])\n",
    "                queue.append(current_node.left)\n",
    "            i += 1\n",
    "\n",
    "            if i < len(values) and values[i] is not None:\n",
    "                current_node.right = Node(val=values[i])\n",
    "                queue.append(current_node.right)\n",
    "            i += 1\n",
    "\n",
    "        return root\n",
    "\n",
    "    def print_level_order(self):\n",
    "        \"\"\"\n",
    "        Prints the level order traversal of the tree (BFS).\n",
    "\n",
    "        Returns:\n",
    "        list: A list of lists, where each sublist contains the values of the tree nodes\n",
    "              at that depth.\n",
    "        \"\"\"\n",
    "        queue = deque([(self, 0)])\n",
    "        result = []\n",
    "        current_level = []\n",
    "        level_number = 0\n",
    "\n",
    "        while queue:\n",
    "            current_node, node_level = queue.popleft()\n",
    "\n",
    "            if node_level > level_number:\n",
    "                result.append(current_level)\n",
    "                current_level = []\n",
    "                level_number = node_level\n",
    "\n",
    "            if current_node:\n",
    "                current_level.append(current_node.val)\n",
    "                queue.append((current_node.left, node_level + 1))\n",
    "                queue.append((current_node.right, node_level + 1))\n",
    "            else:\n",
    "                current_level.append(\"None\")\n",
    "\n",
    "        if current_level:\n",
    "            result.append(current_level)\n",
    "\n",
    "        # Remove trailing \"None\" values\n",
    "        result = [level for level in result if any(\n",
    "            elem != \"None\" for elem in level)]\n",
    "\n",
    "        return result\n",
    "\n",
    "    def print_tree_by_levels(self):\n",
    "        while self:\n",
    "            current = self\n",
    "\n",
    "            while current:\n",
    "                print(current.val, end=\" -> \")\n",
    "                current = current.next\n",
    "\n",
    "            print(\"#\")  # indicate the end of the current level\n",
    "            if self.left:\n",
    "                self = self.left\n",
    "            elif self.right:\n",
    "                self = self.right\n",
    "            else:\n",
    "                # If the current level is completely traversed and there's no left or right child,\n",
    "                # we need to find the next available node in the next level if it exists\n",
    "                next_level = self.next\n",
    "                while next_level and not (next_level.left or next_level.right):\n",
    "                    next_level = next_level.next\n",
    "                self = next_level.left if next_level and next_level.left else next_level.right if next_level else None"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Solution:\n",
    "    def connect(self, root: Node) -> Node:\n",
    "        if not root:\n",
    "            return None\n",
    "\n",
    "        # start with the root node; the head of the first level\n",
    "        level_start = root\n",
    "\n",
    "        # while there are nodes at the current level\n",
    "        while level_start:\n",
    "            current = level_start\n",
    "            # use a dummy node to establish the head of the next level\n",
    "            dummy = Node()\n",
    "            tail = dummy\n",
    "\n",
    "            # while there are nodes at the current level\n",
    "            while current:\n",
    "                # if the current node has a left child, append it to the tail\n",
    "                if current.left:\n",
    "                    tail.next = current.left\n",
    "                    tail = tail.next\n",
    "                # if the current node has a right child, append it to the tail\n",
    "                if current.right:\n",
    "                    tail.next = current.right\n",
    "                    tail = tail.next\n",
    "                current = current.next\n",
    "            # move to the next level and repeat\n",
    "            level_start = dummy.next\n",
    "\n",
    "        return root"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1 -> #\n",
      "2 -> 3 -> #\n",
      "4 -> 5 -> 7 -> #\n"
     ]
    }
   ],
   "source": [
    "# Test\n",
    "s = Solution()\n",
    "\n",
    "root = Node.from_list([1, 2, 3, 4, 5, None, 7])\n",
    "root = s.connect(root)\n",
    "\n",
    "root.print_tree_by_levels()"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
